from typing import Union

from autogen import ConversableAgent

from fastagency.api.openapi.client import OpenAPI
from fastagency.api.openapi.security import APIKeyHeader
from fastagency.runtimes.autogen.autogen import Toolable

WHATSAPP_OPENAPI_URL = "https://dev.infobip.com/openapi/products/whatsapp.json"
WHATSAPP_API_SERVER = "https://api.infobip.com"
WHATSAPP_FUNCTIONS = ["send_whatsapp_text_message"]


class WhatsAppTool(Toolable):
    def __init__(
        self,
        whatsapp_api_key: str,
        whatsapp_openapi_url: str = WHATSAPP_OPENAPI_URL,
        whatsapp_api_server: str = WHATSAPP_API_SERVER,
    ):
        """Create a new WhatsAppTool instance.

        Args:
            whatsapp_api_key (str): The WhatsApp API key.
            whatsapp_openapi_url (str): Url of the openapi schema for Infobip WhatsApp API, defaults to https://dev.infobip.com/openapi/products/whatsapp.json
            whatsapp_api_server (str): Url of the Infobip WhatsApp API server, defaults to https://api.infobip.com
        """
        self.whatsapp_api = OpenAPI.create(
            openapi_url=whatsapp_openapi_url,
            servers=[{"url": whatsapp_api_server}],
        )

        header_authorization = f"App {whatsapp_api_key}"
        self.whatsapp_api.set_security_params(
            APIKeyHeader.Parameters(value=header_authorization)
        )

    def register(
        self,
        *,
        caller: ConversableAgent,
        executor: Union[ConversableAgent, list[ConversableAgent]],
    ) -> None:
        executors = executor if isinstance(executor, list) else [executor]

        self.whatsapp_api._register_for_llm(caller, functions=WHATSAPP_FUNCTIONS)

        for executor in executors:
            self.whatsapp_api._register_for_execution(
                executor, functions=WHATSAPP_FUNCTIONS
            )



from typing import Any, Protocol

import mesop as me

__all__ = ["AuthProtocol"]


class AuthProtocol(Protocol):
    def create_security_policy(
        self, policy: me.SecurityPolicy
    ) -> me.SecurityPolicy: ...

    # maybe me.Component is wrong
    def auth_component(self) -> me.component: ...

    def is_authorized(self, *args: Any, **kwargs: Any) -> bool: ...



import os
import typing
from typing import Any, Callable, Literal, Union

import firebase_admin
import mesop as me
import mesop.labs as mel
from firebase_admin import auth

from ...data_model import State
from ...styles import MesopHomePageStyles
from .firebase_auth_component import FirebaseConfig, firebase_auth_component

__all__ = ["FirebaseAuth"]

if typing.TYPE_CHECKING:
    from ..auth import AuthProtocol


# Avoid re-initializing firebase app (useful for avoiding warning message because of hot reloads).
if firebase_admin._DEFAULT_APP_NAME not in firebase_admin._apps:
    default_app = firebase_admin.initialize_app()


class FirebaseAuth:  # implements AuthProtocol
    def __init__(
        self,
        sign_in_methods: list[Literal["google"]],
        config: FirebaseConfig,
        allowed_users: Union[
            list[str], Callable[[dict[str, Any]], bool], Literal["all"]
        ],  # for callable -> pass the whole decoded token (dict)
    ) -> None:
        """Initialize the Firebase Auth provider.

        Args:
            sign_in_methods: List of authentication methods to enable.
                Currently only supports ["google"].
            config: Firebase configuration containing project settings.
            allowed_users: Specifies user access control:
                - List[str]: List of allowed email addresses
                - Callable: Function taking decoded token and returning boolean
                - "all": Allows all authenticated users (default)

        Raises:
            TypeError: If sign_in_methods is not a list
            ValueError: If no sign-in methods specified, unsupported methods provided,
                or GOOGLE_APPLICATION_CREDENTIALS environment variable is missing
        """
        # mypy check if self is AuthProtocol
        _self: AuthProtocol = self

        self.config = config
        self.allowed_users = allowed_users

        # Validate sign_in_methods type
        if not isinstance(sign_in_methods, list):
            raise TypeError(
                "sign_in_methods must be a list. Example: sign_in_methods=['google']"
            )

        # 2. Remove duplicates
        self.sign_in_methods = list(set(sign_in_methods))

        # 3. Validate sign-in methods
        if not self.sign_in_methods:
            raise ValueError("At least one sign-in method must be specified")

        unsupported_methods = [
            method for method in self.sign_in_methods if method != "google"
        ]
        if unsupported_methods:
            raise ValueError(
                f"Unsupported sign-in method(s): {unsupported_methods}. Currently, only 'google' sign-in is supported."
            )

        if not os.getenv("GOOGLE_APPLICATION_CREDENTIALS"):
            raise ValueError(
                "Error: A service account key is required. Please create one and set the JSON key file path in the `GOOGLE_APPLICATION_CREDENTIALS` environment variable. For more information: https://firebase.google.com/docs/admin/setup#initialize_the_sdk_in_non-google_environments"
            )

    def create_security_policy(self, policy: me.SecurityPolicy) -> me.SecurityPolicy:
        return me.SecurityPolicy(
            dangerously_disable_trusted_types=True,
            allowed_connect_srcs=list(
                set(policy.allowed_connect_srcs or []) | {"*.googleapis.com"}
            ),
            allowed_script_srcs=list(
                set(policy.allowed_script_srcs or [])
                | {
                    "*.google.com",
                    "https://www.gstatic.com",
                    "https://cdn.jsdelivr.net",
                }
            ),
        )

    def is_authorized(self, token: dict[str, Any]) -> bool:
        """Check if the user is authorized based on the token and allowed_users configuration.

        Args:
            token: The decoded Firebase JWT token containing user information.
                Must include an 'email' field for validation.

        Returns:
            bool: True if the user is authorized, False otherwise.

        Raises:
            TypeError: If allowed_users is not of type str, list, or Callable.
            ValueError: If email field is missing in the Firebase token.
        """
        # Check if the email is present in token
        email = token.get("email")
        if not email:
            raise ValueError(
                "Invalid response from Firebase: "
                "`email` field is missing in the token"
            )

        # Handle string-based configuration ("all" or single email)
        if isinstance(self.allowed_users, str):
            if self.allowed_users == "all":
                return True
            return email == self.allowed_users

        # Handle list of allowed email addresses
        if isinstance(self.allowed_users, list):
            return email in {
                addr.strip() if isinstance(addr, str) else addr
                for addr in self.allowed_users
            }

        # Handle custom validation function
        if callable(self.allowed_users):
            return self.allowed_users(token)

        raise TypeError(
            "allowed_users must be one of: "
            "str ('all' or email), "
            "list of emails, "
            "or callable taking token dict"
        )

    def on_auth_changed(self, e: mel.WebEvent) -> None:
        state = me.state(State)
        firebase_auth_token = e.value

        if not firebase_auth_token:
            state.authenticated_user = ""
            return

        decoded_token = auth.verify_id_token(firebase_auth_token)

        if not self.is_authorized(decoded_token):
            raise me.MesopUserException(
                "You are not authorized to access this application. "
                "Please contact the application administrators for access."
            )

        state.authenticated_user = decoded_token["email"]

    # maybe me.Component is wrong
    def auth_component(self) -> me.component:
        styles = MesopHomePageStyles()
        state = me.state(State)
        if state.authenticated_user:
            with me.box(style=styles.logout_btn_container):
                firebase_auth_component(
                    on_auth_changed=self.on_auth_changed, config=self.config
                )
        else:
            with me.box(style=styles.login_box):  # noqa: SIM117
                with me.box(style=styles.login_btn_container):
                    me.text("Sign in to your account", style=styles.header_text)
                    firebase_auth_component(
                        on_auth_changed=self.on_auth_changed, config=self.config
                    )



from typing import Any, Union

from autogen import AssistantAgent, ConversableAgent

from ..tools import WhatsAppTool

WHATSAPP_SYSTEM_MESSAGE = """You are an agent in charge of communication with the WhatsAPP API.

When sending the message, the Body must use the following format:

{{
    "from": "{sender}",
    "to": "receiverNumber",
    "messageId": "test-message-randomInt",
    "content": {{
        "text": "message"
    }},
    "callbackData": "Callback data"
}}

"from" number is always the same.
"""


class WhatsAppAgent(AssistantAgent):  # type: ignore[misc]
    def __init__(
        self,
        *args: Any,
        name: str,
        llm_config: dict[str, Any],
        executor: Union[ConversableAgent, list[ConversableAgent]],
        sender: str,
        whatsapp_api_key: str,
        **kwargs: Any,
    ):
        """Initialize the WhatsAppAgent.

        Args:
            *args (Any): The positional arguments.
            name (str): The name of the agent.
            llm_config (dict[str, Any]): The LLM configuration.
            executor (Union[ConversableAgent, list[ConversableAgent]]): The executor agent(s).
            sender (str): Number of the sender for WhatsApp API.
            whatsapp_api_key (str): The WhatsApp API key
            **kwargs (Any): The keyword arguments.
        """
        super().__init__(
            *args,
            name=name,
            system_message=WHATSAPP_SYSTEM_MESSAGE.format(sender=sender),
            llm_config=llm_config,
            **kwargs,
        )
        self.whatsapp_tool = WhatsAppTool(
            whatsapp_api_key=whatsapp_api_key,
        )

        self.whatsapp_tool.register(caller=self, executor=executor)



from typing import Any, Callable, Optional

import mesop.labs as mel

MEL_WEB_COMPONENT_PATH = "/__fast_agency_internal__/javascript/basic_auth_component.js"


@mel.web_component(path=MEL_WEB_COMPONENT_PATH)  # type: ignore[misc]
def basic_auth_component(
    on_auth_changed: Callable[[mel.WebEvent], Any],
    authenticated_user: Optional[str] = None,
) -> Any:
    return mel.insert_web_component(
        name="basic-auth-component",
        events={
            "authChanged": on_auth_changed,
        },
        properties={"authenticatedUser": authenticated_user},
    )



from .....helpers import check_imports

check_imports(["bcrypt"], "basic_auth")

from .basic_auth import BasicAuth  # noqa: E402
from .basic_auth_component import (  # noqa: E402
    basic_auth_component,
)

__all__ = ["BasicAuth", "basic_auth_component"]



import typing

import bcrypt
import mesop as me
import mesop.labs as mel

from ...data_model import State
from ...styles import MesopHomePageStyles
from .basic_auth_component import basic_auth_component

if typing.TYPE_CHECKING:
    from ..auth import AuthProtocol


class BasicAuth:  # implements AuthProtocol
    def __init__(self, allowed_users: dict[str, str]) -> None:
        """Initialize the authentication component with allowed users.

        Args:
        allowed_users (dict[str, str]): A dictionary where the keys are usernames and the values are passwords.

        Initializes:
        _self (AuthProtocol): Ensures the instance conforms to the AuthProtocol interface.
        allowed_users (dict[str, str]): A dictionary where the keys are usernames and the values are hashed passwords.
        """
        # mypy check if self is AuthProtocol
        _self: AuthProtocol = self

        self.allowed_users = allowed_users

    def create_security_policy(self, policy: me.SecurityPolicy) -> me.SecurityPolicy:
        return policy

    def _verify_password(self, password: str, password_hash: str) -> bool:
        """Verify a password against its hash."""
        return bcrypt.checkpw(password.encode("utf-8"), password_hash.encode("utf-8"))

    def is_authorized(self, username: str, password: str) -> bool:
        if username not in self.allowed_users:
            return False

        password_hash = self.allowed_users[username]
        return self._verify_password(password, password_hash)

    def on_auth_changed(self, e: mel.WebEvent) -> None:
        state = me.state(State)

        if e.value is None:
            state.authenticated_user = ""
            return

        username, password = e.value["username"], e.value["password"]

        if not self.is_authorized(username, password):
            raise me.MesopUserException("Invalid username or password")

        state.authenticated_user = username

    # maybe me.Component is wrong
    def auth_component(self) -> me.component:
        styles = MesopHomePageStyles()
        state = me.state(State)
        if state.authenticated_user:
            with me.box(style=styles.logout_btn_container):
                basic_auth_component(
                    on_auth_changed=self.on_auth_changed,
                    authenticated_user=state.authenticated_user,
                )
        else:
            with me.box(style=styles.login_box):  # noqa: SIM117
                with me.box(style=styles.login_btn_container):
                    basic_auth_component(
                        on_auth_changed=self.on_auth_changed,
                        authenticated_user=state.authenticated_user,
                    )



from typing import Union

from datamodel_code_generator.imports import (
    IMPORT_LITERAL,
    IMPORT_LITERAL_BACKPORT,
    Imports,
)
from datamodel_code_generator.model import pydantic as pydantic_model
from datamodel_code_generator.model import pydantic_v2 as pydantic_model_v2
from datamodel_code_generator.model.base import (
    DataModel,
)
from datamodel_code_generator.reference import Reference

# from datamodel_code_generator.parser import base
from fastapi_code_generator.parser import OpenAPIParser

from ...logging import get_logger

logger = get_logger(__name__)

# Save the original method before patching
original_apply_discriminator_type = OpenAPIParser._Parser__apply_discriminator_type


def patch_apply_discriminator_type() -> None:  # noqa: C901
    def __apply_discriminator_type_patched(  # noqa: C901
        self: OpenAPIParser,
        models: list[DataModel],
        imports: Imports,
    ) -> None:
        for model in models:
            for field in model.fields:
                discriminator = field.extras.get("discriminator")
                if not discriminator or not isinstance(discriminator, dict):
                    continue
                property_name = discriminator.get("propertyName")
                if not property_name:  # pragma: no cover
                    continue
                mapping = discriminator.get("mapping", {})
                for data_type in field.data_type.data_types:
                    if not data_type.reference:  # pragma: no cover
                        continue
                    discriminator_model = data_type.reference.source

                    if not isinstance(  # pragma: no cover
                        discriminator_model,
                        (pydantic_model.BaseModel, pydantic_model_v2.BaseModel),
                    ):
                        continue  # pragma: no cover

                    type_names: list[str] = []

                    def check_paths(
                        model: Union[
                            pydantic_model.BaseModel,
                            pydantic_model_v2.BaseModel,
                            Reference,
                        ],
                        mapping: dict[str, str],
                        type_names: list[str] = type_names,
                    ) -> None:
                        """Helper function to validate paths for a given model."""
                        for name, path in mapping.items():
                            if (
                                model.path.split("#/")[-1] != path.split("#/")[-1]
                            ) and (
                                path.startswith("#/")
                                or model.path[:-1] != path.split("/")[-1]
                            ):
                                t_path = path[str(path).find("/") + 1 :]
                                t_disc = model.path[: str(model.path).find("#")].lstrip(  # noqa: B005
                                    "../"
                                )
                                t_disc_2 = "/".join(t_disc.split("/")[1:])
                                if t_path != t_disc and t_path != t_disc_2:
                                    continue
                            type_names.append(name)

                    # Check the main discriminator model path
                    if mapping:
                        check_paths(discriminator_model, mapping)

                        # Check the base_classes if they exist
                        for base_class in discriminator_model.base_classes:
                            if base_class.reference and base_class.reference.path:
                                check_paths(base_class.reference, mapping)
                    else:
                        type_names = [discriminator_model.path.split("/")[-1]]
                    if not type_names:  # pragma: no cover
                        raise RuntimeError(
                            f"Discriminator type is not found. {data_type.reference.path}"
                        )
                    has_one_literal = False
                    for discriminator_field in discriminator_model.fields:
                        if (
                            discriminator_field.original_name
                            or discriminator_field.name
                        ) != property_name:
                            continue
                        literals = discriminator_field.data_type.literals
                        if (
                            len(literals) == 1 and literals[0] == type_names[0]
                            if type_names
                            else None
                        ):
                            has_one_literal = True
                            continue
                        for (
                            field_data_type
                        ) in discriminator_field.data_type.all_data_types:
                            if field_data_type.reference:  # pragma: no cover
                                field_data_type.remove_reference()
                        discriminator_field.data_type = self.data_type(
                            literals=type_names
                        )
                        discriminator_field.data_type.parent = discriminator_field
                        discriminator_field.required = True
                        imports.append(discriminator_field.imports)
                        has_one_literal = True
                    if not has_one_literal:
                        discriminator_model.fields.append(
                            self.data_model_field_type(
                                name=property_name,
                                data_type=self.data_type(literals=type_names),
                                required=True,
                            )
                        )
                    literal = (
                        IMPORT_LITERAL
                        if self.target_python_version.has_literal_type
                        else IMPORT_LITERAL_BACKPORT
                    )
                    has_imported_literal = any(
                        literal == import_  # type: ignore [comparison-overlap]
                        for import_ in imports
                    )
                    if has_imported_literal:  # pragma: no cover
                        imports.append(literal)

    # Patch the method using the exact mangled name
    OpenAPIParser._Parser__apply_discriminator_type = __apply_discriminator_type_patched

    logger.info("Patched Parser.__apply_discriminator_type,")



from dataclasses import dataclass
from typing import Any, Callable

import mesop.labs as mel

__all__ = ["firebase_auth_component"]

MEL_WEB_COMPONENT_PATH = (
    "/__fast_agency_internal__/javascript/firebase_auth_component.js"
)


@dataclass
class FirebaseConfig:
    api_key: str
    auth_domain: str
    project_id: str
    storage_bucket: str
    messaging_sender_id: str
    app_id: str


@mel.web_component(path=MEL_WEB_COMPONENT_PATH)  # type: ignore[misc]
def firebase_auth_component(
    on_auth_changed: Callable[[mel.WebEvent], Any], config: FirebaseConfig
) -> Any:
    return mel.insert_web_component(
        name="firebase-auth-component",
        events={
            "authChanged": on_auth_changed,
        },
        properties={
            "apiKey": config.api_key,
            "authDomain": config.auth_domain,
            "projectId": config.project_id,
            "storageBucket": config.storage_bucket,
            "messagingSenderId": config.messaging_sender_id,
            "appId": config.app_id,
        },
    )



from .....helpers import check_imports

check_imports(["firebase_admin"], "firebase")

from .firebase_auth import FirebaseAuth  # noqa: E402
from .firebase_auth_component import (  # noqa: E402
    FirebaseConfig,
    firebase_auth_component,
)

__all__ = ["FirebaseAuth", "FirebaseConfig", "firebase_auth_component"]



import subprocess  # nosec B404
from logging import getLogger
from typing import Annotated, Optional

import typer

from .logging import setup_logging

docker_app = typer.Typer(rich_markup_mode="rich")


setup_logging()
logger = getLogger(__name__)


@docker_app.command(
    context_settings={"allow_extra_args": True, "ignore_unknown_options": True},
    help="Build a Docker image for the FastAgency app",
)
def build(
    build_context: Annotated[
        str,
        typer.Argument(
            ...,
            help="Docker build context",
        ),
    ] = ".",
    *,
    file: Annotated[
        str,
        typer.Option(
            "--file",
            "-f",
            help="Name of the Dockerfile",
        ),
    ] = "docker/Dockerfile",
    tag: Annotated[
        str,
        typer.Option(
            "--tag",
            "-t",
            help='Name and optionally a tag (format: "name:tag")',
        ),
    ] = "deploy_fastagency",
    progress: Annotated[
        str,
        typer.Option(
            "--progress",
            help="Set type of progress output (auto, plain, tty, rawjson).",
        ),
    ] = "plain",
    ctx: typer.Context,
) -> None:
    command = [
        "docker",
        "build",
        "-t",
        tag,
        "-f",
        file,
        "--progress",
        progress,
        build_context,
    ]
    command += ctx.args
    typer.echo(
        f"Building FastAgency Docker image with the command: {' '.join(command)}"
    )
    try:
        # Run the docker build command
        result = subprocess.run(  # nosec B603
            command, check=True, capture_output=True, text=True
        )
        typer.echo(result.stdout)
        typer.echo(f"Image '{tag}' built successfully!")
    except subprocess.CalledProcessError as e:
        typer.echo(f"Error: {e.stderr}", err=True)
        raise typer.Exit(code=1) from e


@docker_app.command(
    context_settings={"allow_extra_args": True, "ignore_unknown_options": True},
    help="Run a Docker container for the FastAgency app",
)
def run(
    image: Annotated[
        str,
        typer.Argument(
            ...,
            help="The Docker image to run",
        ),
    ] = "deploy_fastagency",
    *,
    name: Annotated[
        str,
        typer.Option(
            "--name",
            help="Assign a name to the container",
        ),
    ] = "deploy_fastagency",
    env: Annotated[
        Optional[list[str]],
        typer.Option(
            "--env",
            "-e",
            help="Set environment variables",
            show_default=False,
        ),
    ] = None,
    publish: Annotated[
        Optional[list[str]],
        typer.Option(
            "--publish",
            "-p",
            help="Publish a container's port(s) to the host",
            show_default=False,
        ),
    ] = None,
    remove: Annotated[
        bool,
        typer.Option(
            "--rm",
            help="Automatically remove the container and its associated anonymous volumes when it exits",
            is_flag=True,
        ),
    ] = False,
    detach: Annotated[
        bool,
        typer.Option(
            "--detach",
            "-d",
            help="Run container in background and print container ID",
            is_flag=True,
        ),
    ] = True,
    network: Annotated[
        Optional[str],
        typer.Option(
            "--network",
            help="Connect a container to a network",
            show_default=False,
        ),
    ] = None,
    ctx: typer.Context,
) -> None:
    # Construct the docker run command using the provided options
    command = ["docker", "run", "--name", name]

    if env:
        for env_var in env:
            command.extend(["--env", env_var])

    if publish:
        for port in publish:
            command.extend(["--publish", port])
        if "8888:8888" not in publish:
            command.extend(["--publish", "8888:8888"])
    else:
        command.extend(["--publish", "8888:8888"])

    if remove:
        command.append("--rm")

    if detach:
        command.append("--detach")

    if network:
        command.extend(["--network", network])

    command += ctx.args
    command.append(image)

    try:
        typer.echo(
            f"Running FastAgency Docker image with the command: {' '.join(command)}"
        )
        # Run the docker command
        result = subprocess.run(  # nosec B603
            command, check=True, capture_output=True, text=True
        )
        typer.echo(result.stdout)
    except subprocess.CalledProcessError as e:
        typer.echo(f"Error: {e.stderr}", err=True)
        raise typer.Exit(code=1) from e


@docker_app.command(
    context_settings={"allow_extra_args": False, "ignore_unknown_options": False},
    help="Deploy the Docker container for the FastAgency app to Fly.io",
)
def deploy(
    config_file: Annotated[
        str,
        typer.Argument(
            ...,
            help="The Fly.io configuration file",
        ),
    ] = "fly.toml",
    *,
    openai_api_key: Annotated[
        str,
        typer.Option(
            "--openai-api-key",
            help="OpenAI API key",
            envvar="OPENAI_API_KEY",
            show_default=False,
        ),
    ],
    # ctx: typer.Context,
) -> None:
    launch_command = [
        "fly",
        "launch",
        "--config",
        config_file,
        "--copy-config",
        "--yes",
    ]
    # launch_command += ctx.args

    set_secret_command = ["fly", "secrets", "set", "OPENAI_API_KEY=" + openai_api_key]
    try:
        typer.echo(
            f"Deploying FastAgency Docker image to Fly.io with the command: {' '.join(launch_command)}"
        )
        # Run the fly deploy command
        deploy_result = subprocess.run(  # nosec B603
            launch_command, check=True, capture_output=True, text=True
        )
        typer.echo(deploy_result.stdout)

        typer.echo(
            f"Setting OpenAI API key with the command: {' '.join(set_secret_command)}"
        )
        # Run the fly secrets set command
        set_secret_result = subprocess.run(  # nosec B603
            set_secret_command, check=True, capture_output=True, text=True
        )
        typer.echo(set_secret_result.stdout)
        typer.echo("Deployed FastAgency Docker image to Fly.io successfully!")
    except subprocess.CalledProcessError as e:
        typer.echo(f"Error: {e.stderr}", err=True)
        raise typer.Exit(code=1) from e
